# Copyright (C) 2015-2017 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Math utils. Import as "labmath" to prevent conflicts with system math package.
"""
# Use absolute paths for imports so as to prevent a conflict with the
# system "math" module.
from __future__ import absolute_import

# Use floating point division by default.
from __future__ import division

import math
import sys

import numpy as np
import scipy
from scipy import stats

import labm8 as lab
from labm8 import modules

if lab.is_python3():
    from functools import reduce


def ceil(number):
    """
    Return the ceiling of a number as an int.

    This is the smallest integral value >= number.

    Example:

        >>> labmath.ceil(1.5)
        2

    Arguments:

        number (float): A numeric value.

    Returns:

        int: Smallest integer >= number.

    Raises:

        TypeError: If argument is not a numeric value.
    """
    return int(math.ceil(number))


def floor(number):
    """
    Return the floor of a number as an int.

    This is the largest integral value <= number.

    Example:

        >>> labmath.floor(1.5)
        1

    Arguments:

        number (float): A numeric value.

    Returns:

        int: Largest integer <= number.

    Raises:

        TypeError: If argument is not a numeric value.
    """
    return int(math.floor(number))


def sqrt(number):
    """
    Return the square root of a number.
    """
    return math.sqrt(number)


def mean(array):
    """
    Return the mean value of a list of divisible numbers.
    """
    n = len(array)

    if n < 1:
        return 0
    elif n == 1:
        return array[0]
    return sum(array) / n


def geomean(array):
    """
    Return the mean value of a list of divisible numbers.
    """
    n = len(array)

    if n < 1:
        return 0
    elif n == 1:
        return array[0]
    return stats.mstats.gmean(array)


def median(array):
    """
    Return the median value of a list of numbers.
    """
    n = len(array)

    if n < 1:
        return 0
    elif n == 1:
        return array[0]

    sorted_vals = sorted(array)
    midpoint = int(n / 2)
    if n % 2 == 1:
        return sorted_vals[midpoint]
    else:
        return (sorted_vals[midpoint - 1] + sorted_vals[midpoint]) / 2.0


def range(array):
    """
    Return the range between min and max values.
    """
    if len(array) < 1:
        return 0
    return max(array) - min(array)


def variance(array):
    """
    Return the variance of a list of divisible numbers.
    """
    if len(array) < 2:
        return 0
    u = mean(array)
    return sum([(x - u) ** 2 for x in array]) / (len(array) - 1)


def stdev(array):
    """
    Return the standard deviation of a list of divisible numbers.
    """
    return sqrt(variance(array))


def iqr(array, lower, upper):
    """
    Return interquartile range for given array.

    Arguments:

        lower (float): Lower bound for IQR, in range 0 <= lower <= 1.
        upper (float): Upper bound for IQR, in range 0 <= upper <= 1.

    Returns:

        (float, float): Lower and upper IQR values.
    """
    return np.percentile(array, [lower * 100, upper * 100])


def filter_iqr(array, lower, upper):
    """
    Return elements which falls within specified interquartile range.

    Arguments:

        array (list): Sequence of numbers.
        lower (float): Lower bound for IQR, in range 0 <= lower <= 1.
        upper (float): Upper bound for IQR, in range 0 <= upper <= 1.

    Returns:

        list: Copy of original list, with elements outside of IQR
          removed.
    """
    upper, lower = iqr(array, upper, lower)

    new = list(array)
    for x in new[:]:
        if x < lower or x > upper:
            new.remove(x)

    return new


def confinterval(array, conf=0.95, normal_threshold=30, error_only=False,
                 array_mean=None):
    """
    Return the confidence interval of a list for a given confidence.

    Arguments:

        array (list): Sequence of numbers.
        conf (float): Confidence interval, in range 0 <= ci <= 1
        normal_threshold (int): The number of elements in the array is
          < normal_threshold, use a T-distribution. If the number of
          elements in the array is >= normal_threshold, use a normal
          distribution.
        error_only (bool, optional): If true, return only the size of
          symmetrical confidence interval, equal to ci_upper - mean.
        array_mean (float, optional): Optimisation trick for if you
          already know the arithmetic mean of the array to prevent
          this function from re-calculating.

    Returns:

        (float, float): Lower and upper bounds on confidence interval,
          respectively.
    """
    n = len(array)

    if n < 1:
        # We have no data.
        array_mean, c0, c1 = 0, 0, 0
    elif n == 1:
        # We have only a single datapoint, so return that value.
        array_mean, c0, c1 = array[0], array[0], array[0]
    else:
        scale = stdev(array) / sqrt(n)
        # Check if all values are the same.
        values_all_same = all(x == array[0] for x in array[1:])

        if values_all_same:
            # If values are all the same, return that value.
            array_mean, c0, c1 = array[0], array[0], array[0]
        else:
            if array_mean is None: array_mean = mean(array)
            if n < normal_threshold:
                # We have a "small" number of datapoints, so use a
                # t-distribution.
                c0, c1 = scipy.stats.t.interval(conf, n - 1, loc=array_mean,
                                                scale=scale)
            else:
                # We have a "large" number of datapoints, so use a
                # normal distribution.
                c0, c1 = scipy.stats.norm.interval(conf, loc=array_mean,
                                                   scale=scale)

    if error_only:
        return c1 - array_mean
    else:
        return c0, c1



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Extended SQL database interface.
"""
import atexit
import csv
import re
import six
import sqlite3 as sql

import labm8 as lab
from labm8 import fs
from labm8 import io

if lab.is_python3():
    from io import StringIO
else:
    from StringIO import StringIO


class Error(Exception):
    """
    Module-level base error class.
    """
    pass


class SchemaError(Error):
    """
    Error thrown in case of conflicting schemas.
    """
    pass


def placeholders(*columns):
    """
    Generate placeholders string for given arguments.

    Examples:

        >>> placeholders(a, b, c)
        "(?,?,?)"

        >>> placeholders(*sequence)
        "(?,?,?,?,?,?,?,?,?,?)"

        >>> db.execute("INSERT INTO foo VALUES " +
                       placeholders(a, b, c), (a, b, c))

        >>> db.execute("INSERT INTO bar VALUES " +
                       placeholders(*sequence), sequence)

    Arguments:

        *columns: Arguments to generate placeholders for.

    Returns:

        str: Bracketed, comma separated placeholders for given arguments.
    """
    return "(" + ",".join(["?"] * len(columns)) + ")"


def where(*columns):
    """
    String together multiple where clauses.

    Examples:

        >>> where("a", "b", "c")
        "a=? AND b=? AND c=?"

        >>> db.execute("SELECT * FROM foo WHERE {}"
                       .format(where("a", "b", "c")),
                       (a, b, c))

    Arguments:

        *columns: Arguments to generate '<name>=?' placeholders for.

    Returns:

        str: Equivalence checks for all columns.
    """
    return " AND ".join(["{}=?".format(column) for column in columns])


class Database(object):
    def __init__(self, path, tables={}, enable_traces=True):
        """
        Arguments:
            path (str): The path to the database file.
            tables (dictionary of {str: tuple of str}, optional): A diction
              of {name: schema} pairs, where a schema is list of tuple pairs,
              of the form: (name, type).
           enable_traces(bool, optional): Enable traces for user
             defined functions and aggregates.
        """
        self.path = fs.path(path)

        # Create directory if needed.
        parent_dir = fs.dirname(path)
        if parent_dir:
            fs.mkdir(parent_dir)

        self.connection = sql.connect(self.path)

        for name,schema in six.iteritems(tables):
            self.create_table(name, schema)

        io.debug("Opened connection to '{0}'".format(self.path))

        # Register exit handler
        atexit.register(self.close)

        # Enable traces for user defined functions and aggregates. See:
        #
        # https://docs.python.org/2/library/sqlite3.html#sqlite3.enable_callback_tracebacks
        if enable_traces:
            sql.enable_callback_tracebacks(True)

    def __repr__(self):
        return self.path

    def __str__(self):
        return self.__repr__()

    @property
    def tables(self):
        """
        Returns a list of table names.

        Example:

            >>> db.tables
            ["bar", "foo"]

        Returns:

            list of str: One string for each table name.
        """
        select = ("SELECT name FROM sqlite_master",)
        query = self.execute(*select)
        result = query.fetchall()

        # Filter first column from rows.
        return [row[0] for row in result]

    @property
    def schema(self):
        """
        Returns the schema of all tables.

        For each table, return the name, and a list of tuples
        representing the columns. Each column tuple consists of a
        (name, type) pair. Note that additional metadata, such as
        whether a column may be null, or whether a column is a primary
        key, is not returned.

        Example:

            >>> db.schema
            [("bar", (("id", "integer"), ("name", "table"))]

        Returns:

            list of tuples: Each tuple has the format (name, columns), where
              "columns" is a list of tuples of the form (name, type).
        """
        def _info2columns(info):
            return tuple((column["name"], column["type"]) for column in info)

        def _table2tuple(table):
            return (table, _info2columns(self.table_info(table)))

        return [_table2tuple(table) for table in self.tables]

    def table_info(self, table):
        """
        Returns information about the named table.

        See: https://www.sqlite.org/pragma.html#pragma_table_info

        Example:

            >>> db.table_info("foo")
            [{"name": "id", "type": "integer", "primary key": True,
              "notnull": False, "default_value": None}]

        Arguments:

            name (str): The name of the table to lookup.

        Returns:

            list of dicts: One dict per column. Each dict contains the
              keys: "name", "type", "primary key", "notnull", and
              "default_value".

        Raises:

            sql.OperationalError: If table does not exist.
        """
        def _row2dict(row):
            return {
                "name": row[1],
                "type": row[2].lower(),
                "notnull": row[3] == 1,
                "default_value": row[4],
                "primary_key": row[5] == 1
            }

        if table not in self.tables:
            raise sql.OperationalError("Cannot retrieve information about "
                                       "missing table '{0}'".format(table))

        query = self.execute("PRAGMA table_info({table})".format(table=table))
        return [_row2dict(row) for row in query]

    def isempty(self, tables=None):
        """
        Return whether a table or the entire database is empty.

        A database is empty is if it has no tables. A table is empty
        if it has no rows.

        Arguments:

            tables (sequence of str, optional): If provided, check
              that the named tables are empty. If not provided, check
              that all tables are empty.

        Returns:

            bool: True if tables are empty, else false.

        Raises:

            sql.OperationalError: If one or more of the tables do not
              exist.
        """
        tables = tables or self.tables

        for table in tables:
            if self.num_rows(table) > 0:
                return False

        return True

    def num_rows(self, table):
        """
        Return the number of rows in the named table.

        Example:

            >>> db.num_rows("foo")
            3

        Arguments:

            table (str): The name of the table to count the rows in.

        Returns:

            int: The number of rows in the named table.

        Raises:

            sql.OperationalError: If the named table does not exist.
        """
        return self.execute("SELECT Count(*) from " + table).fetchone()[0]

    def close(self):
        """
        Close a database connection.
        """
        self.connection.close()

    def drop_table(self, name):
        """
        Drop an existing table.

        If the table does not exist, nothing happens.

        Arguments:

            name (str): The name of the table to drop.
        """
        if name in self.tables:
            self.execute("DROP TABLE " + name)

    def create_table(self, name, schema):
        """
        Create a new table.

        If the table already exists, nothing happens.

        Example:

            >>> db.create_table("foo", (("id", "integer primary key"),
                                        ("value", "text")))

        Arguments:

           name (str): The name of the table to create.
           schema (sequence of tuples): A list of (name, type) tuples
             representing each of the columns.
        """
        columns = [" ".join(column) for column in schema]
        self.execute("CREATE TABLE IF NOT EXISTS {name} ({columns})"
                     .format(name=name, columns=",".join(columns)))

    def empty_table(self, name):
        """
        Delete all rows in a table.

        If the table does not exist, nothing happens.

        Arguments:

            name (str): The name of the table to empty.
        """
        if name in self.tables:
            self.execute("DELETE FROM " + name)

    def create_table_from(self, name, src):
        """
        Create a new table with same schema as the source.

        If the named table already exists, nothing happens.

        Arguments:

            name (str): The name of the table to create.
            src (str): The name of the source table to duplicate.

        Raises:

            sql.OperationalError: If source table does not exist.
        """
        # Lookup the command which was used to create the "src" table.
        query = self.execute("SELECT sql FROM sqlite_master WHERE "
                             "type='table' and name=?", (src,))
        try:
            cmd = query.fetchone()[0]
        except TypeError:
            raise sql.OperationalError("Cannot copy non-existent table '{0}'"
                                       .format(src))

        # Modify the original command to replace the old table name
        # with the new one.
        new_cmd = re.sub("(CREATE TABLE) \w+", "\\1 " + name, cmd,
                         re.IGNORECASE)

        # Execute this new command.
        self.execute(new_cmd)

    def copy_table(self, src, dst):
        """
        Create a carbon copy of the source table.

        Arguments:

            src (str): The name of the table to copy.
            dst (str): The name of the target duplicate table.

        Raises:

            sql.OperationalError: If source table does not exist.
        """
        # Create table.
        self.create_table_from(dst, src)

        # Copy contents of src to dst.
        self.execute("INSERT INTO {dst} SELECT * FROM {src}"
                     .format(dst=dst, src=src))

        # Commit changes.
        self.commit()

    def execute(self, *args):
        """
        Execute the given arguments.
        """
        return self.connection.cursor().execute(*args)

    def executemany(self, *args):
        """
        Execute the given arguments.
        """
        return self.connection.cursor().executemany(*args)

    def executescript(self, *args):
        """
        Execute the given arguments.
        """
        return self.connection.cursor().executescript(*args)

    def commit(self):
        """
        Commit the current transaction.

        Make sure to call this method after you've modified the
        database's state!
        """
        return self.connection.commit()

    def attach(self, path, name):
        """
        Attach a database.

        Arguments:

            path (str): Path to the database to merge.
            name (str): Name to attach database as.
        """
        self.execute("ATTACH ? as ?", (path, name))

    def detach(self, name):
        """
        Detach a database.

        Arguments:

            name (str): Name of database to detach.
        """
        self.execute("DETACH ?", (name,))

    def export_csv(self, table, output=None, columns="*", **kwargs):
        """
        Export a table to a CSV file.

        If an output path is provided, write to file. Else, return a
        string.

        Wrapper around pandas.sql.to_csv(). See:
        http://pandas.pydata.org/pandas-docs/stable/io.html#io-store-in-csv

        Arguments:

            table (str): Name of the table to export.
            output (str, optional): Path of the file to write.
            columns (str, optional): A comma separated list of columns
              to export.
            **kwargs: Additional args passed to pandas.sql.to_csv()

        Returns:

            str: CSV string, or None if writing to file.

        Raises:

            IOError: In case of error writing to file.
            SchemaError: If the named table is not found.
        """
        import pandas.io.sql as panda

        # Determine if we're writing to a file or returning a string.
        isfile = output is not None
        output = output or StringIO()

        if table not in self.tables:
            raise SchemaError("Cannot find table '{table}'"
                              .format(table=table))

        # Don't print row indexes by default.
        if "index" not in kwargs:
            kwargs["index"] = False

        table = panda.read_sql("SELECT {columns} FROM {table}"
                               .format(columns=columns, table=table),
                               self.connection)
        table.to_csv(output, **kwargs)
        return None if isfile else output.getvalue()



# Copyright (C) 2015-2017 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Wrapper for invoking Makefiles.
"""
import os
import re
import subprocess

import labm8
from labm8 import fs
from labm8 import system


class Error(Exception):
    """
    Module-level error class.
    """
    pass


class MakeError(Error):
    """
    Thrown if make command fails.
    """
    pass


class NoMakefileError(Error):
    """
    Thrown if a directory does not contain a Makefile.
    """
    pass


class NoTargetError(Error):
    """
    Thrown if there is no rule for the requested target.
    """
    pass


_BAD_TARGET_RE = re.compile(r"No rule to make target "
                            "'.+'.  Stop.")


def make(target="all", dir=".", **kwargs):
    """
    Run make.

    Arguments:

        target (str, optional): Name of the target to build. Defaults
          to "all".
        dir (str, optional): Path to directory containing Makefile.
        **kwargs (optional): Any additional arguments to be passed to
          system.run().

    Returns:

        (int, str, str): The first element is the return code of the
          make command. The second and third elements are the stdout
          and stderr of the process.

    Raises:

        NoMakefileError: In case a Makefile is not found in the target
          directory.
        NoTargetError: In case the Makefile does not support the
          requested target.
        MakeError: In case the target rule fails.
    """
    if not fs.isfile(fs.path(dir, "Makefile")):
        raise NoMakefileError("No makefile in '{}'".format(fs.abspath(dir)))

    fs.cd(dir)

    # Default parameters to system.run()
    if "timeout" not in kwargs: kwargs["timeout"] = 300

    ret, out, err = system.run(["make", target], **kwargs)
    fs.cdpop()

    if ret > 0:
        if re.search(_BAD_TARGET_RE, err):
            raise NoTargetError("No rule for target '{}'"
                                .format(target))
        else:
            raise MakeError("Target '{}' failed".format(target))

        raise MakeError("Failed")

    return ret, out, err


def clean(**kwargs):
    """
    Run make clean.

    Equivalent to invoking make() with target="clean".

    Arguments:

        **kwargs (optional): Any additional arguments to be passed to
          make().

    Returns:

        (int, str, str): The first element is the return code of the
          make command. The second and third elements are the stdout
          and stderr of the process.

    Raises:

        NoMakefileError: In case a Makefile is not found in the target
          directory.
        NoTargetError: In case the Makefile does not support the
          requested target.
        MakeError: In case the target rule fails.
    """
    make(target="clean", **kwargs)



# Copyright (C) 2015-2017 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Time utilities.
"""
import datetime

DEFAULT_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def strfmt(datetime, format=DEFAULT_DATETIME_FORMAT):
    """
    Format date to string.
    """
    return datetime.strftime(format)


def now():
    """
    Get the current datetime.
    """
    return datetime.datetime.now()


def nowstr(format=DEFAULT_DATETIME_FORMAT):
    """
    Convenience wrapper to get the current time as a string.

    Equivalent to invoking strfmt(now()).
    """
    return strfmt(now(), format=format)



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
JSON parser which supports comments.
"""
import json
import re

import labm8
from labm8 import fs


def format_json(data):
    """
    Pretty print JSON.

    Arguments:
        data (dict): JSON blob.

    Returns:
        str: Formatted JSON
    """
    return json.dumps(data, sort_keys=True, indent=2, separators=(',', ': '))


def read_file(*components, **kwargs):
    """
    Load a JSON data blob.

    Arguments:
        path (str): Path to file.
        must_exist (bool, otional): If False, return empty dict if file does
            not exist.

    Returns:
        array or dict: JSON data.

    Raises:
        File404: If path does not exist, and must_exist is True.
        InvalidFile: If JSON is malformed.
    """
    must_exist = kwargs.get("must_exist", True)

    if must_exist:
        path = fs.must_exist(*components)
    else:
        path = fs.path(*components)

    try:
        with open(path) as infile:
            return loads(infile.read())
    except ValueError as e:
        raise ValueError(
            "malformed JSON file '{path}'. Message from parser: {err}"
            .format(path=fs.basename(path), err=str(e)))
    except IOError as e:
        if not must_exist:
            return {}
        else:
            return e


def write_file(path, data, format=True):
    """
    Write JSON data to file.

    Arguments:
        path (str): Destination.
        data (dict or list): JSON serializable data.
        format (bool, optional): Pretty-print JSON data.
    """
    if format:
        fs.write_file(path, format_json(data))
    else:
        fs.write_file(path, json.dumps(data))


def loads(text, **kwargs):
    """
    Deserialize `text` (a `str` or `unicode` instance containing a JSON
    document with Python or JavaScript like comments) to a Python object.

    Supported comment types: `// comment` and `# comment`.

    Taken from `commentjson <https://github.com/vaidik/commentjson>`_, written
    by `Vaidik Kapoor <https://github.com/vaidik>`_.

    Copyright (c) 2014 Vaidik Kapoor, MIT license.

    Arguments:
        text (str): serialized JSON string with or without comments.
        **kwargs (optional): all the arguments that
            `json.loads <http://docs.python.org/2/library/json.html#json.loads>`_
            accepts.

    Returns:
        `dict` or `list`: Decoded JSON.
    """
    regex = r'\s*(#|\/{2}).*$'
    regex_inline = r'(:?(?:\s)*([A-Za-z\d\.{}]*)|((?<=\").*\"),?)(?:\s)*(((#|(\/{2})).*)|)$'
    lines = text.split('\n')

    for index, line in enumerate(lines):
        if re.search(regex, line):
            if re.search(r'^' + regex, line, re.IGNORECASE):
                lines[index] = ""
            elif re.search(regex_inline, line):
                lines[index] = re.sub(regex_inline, r'\1', line)

    return json.loads('\n'.join(lines), **kwargs)



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utilities for generating LaTeX.
"""
from __future__ import print_function

import re

import labm8 as lab
from labm8 import fs
from labm8 import io

if lab.is_python3():
    from io import StringIO
else:
    from StringIO import StringIO


class Error(Exception):
    """
    Module-level error.
    """
    pass


def escape(text):
    return re.sub(r"(_)", r"\\\g<1>", str(text))


def wrap_bold(text):
    return "\\textbf{" + text + "}"


def write_table_body(data, output=None, headers=None,
                     header_fmt=wrap_bold, hline_after_header=True,
                     hline_before=False, hline_after=False):
    def _write_row(row):
        output.write(" & ".join(escape(column) for column in row) + "\\\\\n")

    # Determine if we're writing to a file or returning a string.
    isfile = output is not None
    output = output or StringIO()

    # Write hline before body.
    if hline_before:
        output.write("\\hline\n")

    # Write headers.
    if headers:
        _write_row(header_fmt(str(column)) for column in headers)
        # Write hline after headers.
        if hline_after_header:
            output.write("\\hline\n")

    # Write table entries.
    for row in data:
        _write_row(row)

    # Write hline after body.
    if hline_after:
        output.write("\\hline\n")

    return None if isfile else output.getvalue()


def table(rows, columns=None, output=None, data_args={}, **kwargs):
    """
    Return a LaTeX formatted string of "list of list" table data.

    See: http://pandas.pydata.org/pandas-docs/dev/generated/pandas.DataFrame.html

    Requires the "booktabs" package to be included in LaTeX preamble:

        \\usepackage{booktabs}

    Examples:

        >>> fmt.print([("foo", 1), ("bar", 2)])
             0  1
        0  foo  1
        1  bar  2

        >>> fmt.print([("foo", 1), ("bar", 2)], columns=("type", "value"))
          type  value
        0  foo      1
        1  bar      2

    Arguments:

        rows (list of list): Data to format, one row per element,
          multiple columns per row.
        columns (list of str, optional): Column names.
        output (str, optional): Path to output file.
        data_args (dict, optional): Any additional kwargs to pass to
          pandas.DataFrame constructor.
        **kwargs: Any additional arguments to pass to
          pandas.DataFrame.to_latex().

    Returns:

        str: Formatted data as LaTeX table.

    Raises:

        Error: If number of columns (if provided) does not equal
          number of columns in rows; or if number of columns is not
          consistent across all rows.
    """
    import pandas

    # Number of columns.
    num_columns = len(rows[0])

    # Check that each row is the same length.
    for i,row in enumerate(rows[1:]):
        if len(row) != num_columns:
            raise Error("Number of columns in row {i_row} ({c_row}) "
                        "does not match number of columns in row 0 ({z_row})"
                        .format(i_row=i, c_row=len(row), z_row=num_columns))

    # Check that (if supplied), number of columns matches number of
    # columns in rows.
    if columns is not None and len(columns) != num_columns:
        raise Error("Number of columns in header ({c_header}) does not "
                    "match the number of columns in the data ({c_rows})"
                    .format(c_header=len(columns), c_rows=num_columns))

    # Default arguments.
    if "index" not in kwargs:
        kwargs["index"] = False

    data_args["columns"] = columns

    string = pandas.DataFrame(list(rows), **data_args).to_latex(**kwargs)
    if output is None:
        return string
    else:
        print(string, file=open(output, "w"))
        io.info("Wrote", output)



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Python type utilities.
"""
import inspect
import sys

from collections import Mapping
from six import string_types

def is_str(s):
    """
    Return whether variable is string type.

    On python 3, unicode encoding is *not* string type. On python 2, it is.

    Arguments:
        s: Value.

    Returns:
        bool: True if is string, else false.
    """
    return isinstance(s, string_types)


def is_dict(obj):
    """
    Check if an object is a dict.
    """
    return isinstance(obj, dict)


def is_seq(obj):
    """
    Check if an object is a sequence.
    """
    return (not is_str(obj) and not is_dict(obj) and
            (hasattr(obj, "__getitem__") or hasattr(obj, "__iter__")))


def flatten(lists):
    """
    Flatten a list of lists.
    """
    return [item for sublist in lists for item in sublist]


def update(dst, src):
    """
    Recursively update values in dst from src.

    Unlike the builtin dict.update() function, this method will decend into
    nested dicts, updating all nested values.

    Arguments:
        dst (dict): Destination dict.
        src (dict): Source dict.

    Returns:
        dict: dst updated with entries from src.
    """
    for k, v in src.items():
        if isinstance(v, Mapping):
            r = update(dst.get(k, {}), v)
            dst[k] = r
        else:
            dst[k] = src[k]
    return dst


def dict_values(src):
    """
    Recursively get values in dict.

    Unlike the builtin dict.values() function, this method will descend into
    nested dicts, returning all nested values.

    Arguments:
        src (dict): Source dict.

    Returns:
        list: List of values.
    """
    for v in src.values():
        if isinstance(v, dict):
            for v in dict_values(v):
                yield v
        else:
            yield v


def get_class_that_defined_method(meth):
    """
    Return the class that defines a method.

    Arguments:
        meth (str): Class method.

    Returns:
        class: Class object, or None if not a class method.
    """
    if sys.version_info >= (3, 0):
        # Written by @Yoel http://stackoverflow.com/a/25959545
        if inspect.ismethod(meth):
            for cls in inspect.getmro(meth.__self__.__class__):
               if cls.__dict__.get(meth.__name__) is meth:
                    return cls
            meth = meth.__func__ # fallback to __qualname__ parsing
        if inspect.isfunction(meth):
            cls = getattr(inspect.getmodule(meth),
                          meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
            if isinstance(cls, type):
                return cls
    else:
        try:
            # Writted by @Alex Martelli http://stackoverflow.com/a/961057
            for cls in inspect.getmro(meth.im_class):
                if meth.__name__ in cls.__dict__:
                    return cls
        except AttributeError:
            return None
    return None



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tarball util.
"""

import tarfile

from labm8 import fs


def unpack_archive(*components, **kwargs) -> str:
    """
    Unpack a compressed archive.

    Arguments:
        *components (str[]): Absolute path.
        **kwargs (dict, optional): Set "compression" to compression type.
            Default: bz2. Set "dir" to destination directory. Defaults to the
            directory of the archive.

    Returns:
        str: Path to directory.
    """
    path = fs.abspath(*components)
    compression = kwargs.get("compression", "bz2")
    dir = kwargs.get("dir", fs.dirname(path))

    fs.cd(dir)
    tar = tarfile.open(path, "r:" + compression)
    tar.extractall()
    tar.close()
    fs.cdpop()

    return dir



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The fast hash cache.

Checksum directories and cache results. If a directory has not been modified,
subsequent hashes are cache hits. Hashes are recomputed lazily, when a
directory (or any of its subdirectories) have been modified.
"""
import checksumdir
import os
import sqlite3
import time

from labm8 import fs


class DirHashCache(object):
    def __init__(self, path, hash='sha1'):
        """
        Instantiate a directory checksum cache.

        Arguments:
            path (str): Path to persistent cache store.
            hash (str, optional): Hash algorithm to use, e.g. 'md5', 'sha1'.
        """
        self.path = fs.path(path)
        self.hash = hash

        db = sqlite3.connect(self.path)
        c = db.cursor()
        c.execute("""\
CREATE TABLE IF NOT EXISTS dirhashcache (
        path TEXT NOT NULL,
        date DATETIME NOT NULL,
        hash TEXT NOT NULL,
        PRIMARY KEY(path)
)""")
        db.commit()
        db.close()

    def clear(self):
        """
        Remove all cache entries.
        """
        db = sqlite3.connect(self.path)
        c = db.cursor()
        c.execute("DELETE FROM dirhashcache")
        db.commit()
        db.close()

    def dirhash(self, path, **dirhash_opts):
        """
        Compute the hash of a directory.

        Arguments:
           path: Directory.
           **dirhash_opts: Additional options to checksumdir.dirhash().

        Returns:
            str: Checksum of directory.
        """
        path = fs.path(path)
        last_modified = time.ctime(max(
            max(os.path.getmtime(os.path.join(root, file)) for file in files)
            for root,_,files in os.walk(path)))

        db = sqlite3.connect(self.path)
        c = db.cursor()
        c.execute("SELECT date, hash FROM dirhashcache WHERE path=?", (path,))
        cached = c.fetchone()

        if cached:
            cached_date, cached_hash = cached
            if cached_date == last_modified:
                # cache hit
                dirhash = cached_hash
            else:
                # out of date cache
                dirhash = checksumdir.dirhash(path, self.hash, **dirhash_opts)
                c.execute("UPDATE dirhashcache SET date=?, hash=? WHERE path=?",
                          (last_modified, dirhash, path))
                db.commit()
        else:
            # new entry
            dirhash = checksumdir.dirhash(path, self.hash, **dirhash_opts)
            c.execute("INSERT INTO dirhashcache VALUES (?,?,?)",
                      (path, last_modified, dirhash))
            db.commit()

        db.close()
        return dirhash



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Hashing and cryptography utils.
"""
import hashlib


def _checksum(hash_fn, data):
    return hash_fn(data).hexdigest()


def _checksum_str(hash_fn, string, encoding='utf-8'):
    return _checksum(hash_fn, string.encode(encoding))


def _checksum_list(hash_fn, *elems):
    string = "".join(sorted(str(x) for x in elems))
    return _checksum_str(hash_fn, string)


def _checksum_file(hash_fn, path):
    with open(path, 'rb') as infile:
        ret = _checksum(hash_fn, infile.read())
    return ret


def sha1(data):
    """
    Return the sha1 of "data".

    Arguments:
        data (bytes): Data.

    Returns:
        str: Hex encoded.
    """
    return _checksum(hashlib.sha1, data)


def sha1_str(string, encoding='utf-8'):
    """
    Return the sha1 of string "data".

    Arguments:
        string: String.

    Returns:
        str: Hex encoded.
    """
    return _checksum_str(hashlib.sha1, string, encoding=encoding)


def sha1_list(*elems):
    """
    Return the sha1 of all elements of a list.

    Arguments:
        *elems: List of stringifiable data.

    Returns:
        str: Hex encoded.
    """
    return _checksum_list(hashlib.sha1, *elems)


def sha1_file(path):
    """
    Return the sha1 of file at "path".

    Arguments:
        path (str): Path to file

    Returns:
        str: Hex encoded.
    """
    return _checksum_file(hashlib.sha1, path)


def md5(data):
    """
    Return the md5 of "data".

    Arguments:
        data (bytes): Data.

    Returns:
        str: Hex encoded.
    """
    return _checksum(hashlib.md5, data)


def md5_str(string, encoding='utf-8'):
    """
    Return the md5 of string "data".

    Arguments:
        string: String.

    Returns:
        str: Hex encoded.
    """
    return _checksum_str(hashlib.md5, string, encoding=encoding)


def md5_list(*elems):
    """
    Return the md5 of all elements of a list.

    Arguments:
        *elems: List of stringifiable data.

    Returns:
        str: Hex encoded.
    """
    return _checksum_list(hashlib.md5, *elems)


def md5_file(path):
    """
    Return the md5 of file at "path".

    Arguments:
        path (str): Path to file

    Returns:
        str: Hex encoded.
    """
    return _checksum_file(hashlib.md5, path)


def sha256(data):
    """
    Return the sha256 of "data".

    Arguments:
        data (bytes): Data.

    Returns:
        str: Hex encoded.
    """
    return _checksum(hashlib.sha256, data)


def sha256_str(string, encoding='utf-8'):
    """
    Return the sha256 of string "data".

    Arguments:
        string: String.

    Returns:
        str: Hex encoded.
    """
    return _checksum_str(hashlib.sha256, string, encoding=encoding)


def sha256_list(*elems):
    """
    Return the sha256 of all elements of a list.

    Arguments:
        *elems: List of stringifiable data.

    Returns:
        str: Hex encoded.
    """
    return _checksum_list(hashlib.sha256, *elems)


def sha256_file(path):
    """
    Return the sha256 of file at "path".

    Arguments:
        path (str): Path to file

    Returns:
        str: Hex encoded.
    """
    return _checksum_file(hashlib.sha256, path)



# Copyright (C) 2015-2017 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
from setuptools import setup

with open('./requirements.txt') as infile:
    requirements = [x.strip() for x in infile.readlines() if x.strip()]

setup(name="labm8",
      version="0.1.1",
      description="Utils for manipulating quantitative experimental data",
      url="https://github.com/ChrisCummins/labm8",
      author="Chris Cummins",
      author_email="chrisc.101@gmail.com",
      license="GNU General Public License, Version 3",
      packages=["labm8"],
      test_suite="nose.collector",
      tests_require=["nose"],
      install_requires=requirements,
      zip_safe=True)



# Copyright (C) 2015-2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
"""
Profiling API for timing critical paths in code.
"""
from __future__ import absolute_import
from __future__ import print_function

import inspect
import os
import random
import sys

from time import time

import labm8 as lab
from labm8 import fs
from labm8 import io
from labm8 import types


__TIMERS = {}


def is_enabled():
    return os.environ.get("PROFILE") is not None


def enable():
    os.environ["PROFILE"] = "1"


def disable():
    os.environ.pop("PROFILE", None)


def isrunning(name):
    """
    Check if a timer is running.

    Arguments:

        name (str, optional): The name of the timer to check.

    Returns:

        bool: True if timer is running, else False.
    """
    return name in _timers


def start(name):
    """
    Start a new profiling timer.

    Arguments:

        name (str, optional): The name of the timer to create. If no
          name is given, the resulting timer is anonymous. There can
          only be one anonymous timer.
        unique (bool, optional): If true, then ensure that timer name
          is unique. This is to prevent accidentally resetting an
          existing timer.

    Returns:

        bool: Whether or not profiling is enabled.
    """
    if is_enabled():
        __TIMERS[name] = time()
    return is_enabled()


def stop(name, file=sys.stderr):
    """
    Stop a profiling timer.

    Arguments:

        name (str): The name of the timer to stop. If no name is given, stop
            the global anonymous timer.

    Returns:

        bool: Whether or not profiling is enabled.

    Raises:

        KeyError: If the named timer does not exist.
    """
    if is_enabled():
        elapsed = (time() - __TIMERS[name]) * 1000
        del __TIMERS[name]
        print("[prof] {}  {:.1f} ms".format(name, elapsed), file=file)
    return is_enabled()


def profile(fun, *args, **kwargs):
    """
    Profile a function.
    """
    timer_name = kwargs.pop("prof_name", None)

    if not timer_name:
        module = inspect.getmodule(fun)
        c = [module.__name__]
        parentclass = types.get_class_that_defined_method(fun)
        if parentclass:
            c.append(parentclass.__name__)
        c.append(fun.__name__)
        timer_name = ".".join(c)

    start(timer_name)
    ret = fun(*args, **kwargs)
    stop(timer_name)
    return ret


def timers():
    """
    Iterate over all timers.

    Returns:
        Iterable[str]: An iterator over all time names.
    """
    for name in __TIMERS:
        yield name



# Copyright (C) 2015-2017 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utils for handling python modules.
"""
import imp
import sys

import labm8 as lab
from labm8 import io


def import_foreign(name, custom_name=None):
    """
    Import a module with a custom name.

    NOTE this is only needed for Python2. For Python3, import the
    module using the "as" keyword to declare the custom name.

    For implementation details, see:
    http://stackoverflow.com/a/6032023

    Example:

      To import the standard module "math" as "std_math":

          if labm8.is_python3():
            import math as std_math
          else:
            std_math = modules.import_foreign("math", "std_math")

    Arguments:

        name (str): The name of the module to import.
        custom_name (str, optional): The custom name to assign the module to.

    Raises:
        ImportError: If the module is not found.
    """
    if lab.is_python3():
        io.error(("Ignoring attempt to import foreign module '{mod}' "
                  "using python version {major}.{minor}"
                  .format(mod=name, major=sys.version_info[0],
                          minor=sys.version_info[1])))
        return

    custom_name = custom_name or name

    f, pathname, desc = imp.find_module(name, sys.path[1:])
    module = imp.load_module(custom_name, f, pathname, desc)
    f.close()

    return module
